package java.io;

import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;
import java.util.concurrent.ConcurrentHashMap;
import java.lang.ref.ReferenceQueue;
import java.util.concurrent.ConcurrentMap;
import sun.security.action.GetBooleanAction;
import sun.reflect.misc.ReflectUtil;
import java.security.AccessController;
import java.security.PrivilegedAction;

public class ObjectOutputStream extends OutputStream implements ObjectOutput, ObjectStreamConstants
{
    private final ObjectOutputStream.BlockDataOutputStream bout;
    private final ObjectOutputStream.HandleTable handles;
    private final ObjectOutputStream.ReplaceTable subs;
    private int protocol;
    private int depth;
    private byte[] primVals;
    private final boolean enableOverride;
    private boolean enableReplace;
    private SerialCallbackContext curContext;
    private ObjectOutputStream.PutFieldImpl curPut;
    private final ObjectOutputStream.DebugTraceInfoStack debugInfoStack;
    private static final boolean extendedDebugInfo;
    
    public ObjectOutputStream(OutputStream outputStream) throws IOException {
        this.protocol = 2;
        this.verifySubclass();
        this.bout = new BlockDataOutputStream(outputStream);
        this.handles = new ObjectOutputStream.HandleTable(10, 3.0f);
        this.subs = new ObjectOutputStream.ReplaceTable(10, 3.0f);
        this.enableOverride = false;
        this.writeStreamHeader();
        this.bout.setBlockDataMode(true);
        if (ObjectOutputStream.extendedDebugInfo) {
            this.debugInfoStack = new ObjectOutputStream.DebugTraceInfoStack();
        }
        else {
            this.debugInfoStack = null;
        }
    }
    
    protected ObjectOutputStream() throws IOException, SecurityException {
        this.protocol = 2;
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(ObjectOutputStream.SUBCLASS_IMPLEMENTATION_PERMISSION);
        }
        this.bout = null;
        this.handles = null;
        this.subs = null;
        this.enableOverride = true;
        this.debugInfoStack = null;
    }
    
    public void useProtocolVersion(int n) throws IOException {
        if (this.handles.size() != 0) {
            throw new IllegalStateException("stream non-empty");
        }
        switch (n) {
            case 1:
            case 2: {
                this.protocol = n;
                return;
            }
            default: {
                throw new IllegalArgumentException("unknown version: " + n);
            }
        }
    }
    
    @Override
    public final void writeObject(Object o) throws IOException {
        if (this.enableOverride) {
            this.writeObjectOverride(o);
            return;
        }
        try {
            this.writeObject0(o, false);
        }
        catch (IOException ex) {
            if (this.depth == 0) {
                this.writeFatalException(ex);
            }
            throw ex;
        }
        return;
    }
    
    protected void writeObjectOverride(Object o) throws IOException {
    }
    
    public void writeUnshared(Object o) throws IOException {
        try {
            this.writeObject0(o, true);
        }
        catch (IOException ex) {
            if (this.depth == 0) {
                this.writeFatalException(ex);
            }
            throw ex;
        }
    }
    
    public void defaultWriteObject() throws IOException {
        SerialCallbackContext curContext = this.curContext;
        if (curContext == null) {
            throw new NotActiveException("not in call to writeObject");
        }
        Object obj = curContext.getObj();
        ObjectStreamClass desc = curContext.getDesc();
        this.bout.setBlockDataMode(false);
        this.defaultWriteFields(obj, desc);
        this.bout.setBlockDataMode(true);
        return;
    }
    
    public ObjectOutputStream.PutField putFields() throws IOException {
        if (this.curPut == null) {
            SerialCallbackContext curContext = this.curContext;
            if (curContext == null) {
                throw new NotActiveException("not in call to writeObject");
            }
            curContext.getObj();
            this.curPut = new ObjectOutputStream.PutFieldImpl(curContext.getDesc());
        }
        return this.curPut;
    }
    
    public void writeFields() throws IOException {
        if (this.curPut == null) {
            throw new NotActiveException("no current PutField object");
        }
        this.bout.setBlockDataMode(false);
        this.curPut.writeFields();
        this.bout.setBlockDataMode(true);
        return;
    }
    
    public void reset() throws IOException {
        if (this.depth != 0) {
            throw new IOException("stream active");
        }
        this.bout.setBlockDataMode(false);
        this.bout.writeByte(121);
        this.clear();
        this.bout.setBlockDataMode(true);
        return;
    }
    
    protected void annotateClass(Class<?> clazz) throws IOException {
    }
    
    protected void annotateProxyClass(Class<?> clazz) throws IOException {
    }
    
    protected Object replaceObject(Object o) throws IOException {
        return o;
    }
    
    protected boolean enableReplaceObject(boolean enableReplace) throws SecurityException {
        if (enableReplace == this.enableReplace) {
            return enableReplace;
        }
        if (enableReplace) {
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                securityManager.checkPermission(ObjectOutputStream.SUBSTITUTION_PERMISSION);
            }
        }
        this.enableReplace = enableReplace;
        return !this.enableReplace;
    }
    
    protected void writeStreamHeader() throws IOException {
        this.bout.writeShort(-21267);
        this.bout.writeShort(5);
    }
    
    protected void writeClassDescriptor(ObjectStreamClass objectStreamClass) throws IOException {
        objectStreamClass.writeNonProxy(this);
    }
    
    @Override
    public void write(int n) throws IOException {
        this.bout.write(n);
    }
    
    @Override
    public void write(byte[] array) throws IOException {
        this.bout.write(array, 0, array.length, false);
    }
    
    @Override
    public void write(byte[] array, int n, int n2) throws IOException {
        if (array == null) {
            throw new NullPointerException();
        }
        int n3 = n + n2;
        if (n < 0 || n2 < 0 || n3 > array.length || n3 < 0) {
            throw new IndexOutOfBoundsException();
        }
        this.bout.write(array, n, n2, false);
        return;
    }
    
    @Override
    public void flush() throws IOException {
        this.bout.flush();
    }
    
    protected void drain() throws IOException {
        this.bout.drain();
    }
    
    @Override
    public void close() throws IOException {
        this.flush();
        this.clear();
        this.bout.close();
    }
    
    @Override
    public void writeBoolean(boolean b) throws IOException {
        this.bout.writeBoolean(b);
    }
    
    @Override
    public void writeByte(int n) throws IOException {
        this.bout.writeByte(n);
    }
    
    @Override
    public void writeShort(int n) throws IOException {
        this.bout.writeShort(n);
    }
    
    @Override
    public void writeChar(int n) throws IOException {
        this.bout.writeChar(n);
    }
    
    @Override
    public void writeInt(int n) throws IOException {
        this.bout.writeInt(n);
    }
    
    @Override
    public void writeLong(long n) throws IOException {
        this.bout.writeLong(n);
    }
    
    @Override
    public void writeFloat(float n) throws IOException {
        this.bout.writeFloat(n);
    }
    
    @Override
    public void writeDouble(double n) throws IOException {
        this.bout.writeDouble(n);
    }
    
    @Override
    public void writeBytes(String s) throws IOException {
        this.bout.writeBytes(s);
    }
    
    @Override
    public void writeChars(String s) throws IOException {
        this.bout.writeChars(s);
    }
    
    @Override
    public void writeUTF(String s) throws IOException {
        this.bout.writeUTF(s);
    }
    
    int getProtocolVersion() {
        return this.protocol;
    }
    
    void writeTypeString(String s) throws IOException {
        if (s == null) {
            this.writeNull();
        }
        else {
            int lookup;
            if ((lookup = this.handles.lookup(s)) != -1) {
                this.writeHandle(lookup);
            }
            else {
                this.writeString(s, false);
            }
        }
    }
    
    private void verifySubclass() {
        Class<Object> class1 = this.getClass();
        if (class1 == ObjectOutputStream.class) {
            return;
        }
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager == null) {
            return;
        }
        ObjectStreamClass.processQueue(ObjectOutputStream.Caches.subclassAuditsQueue, ObjectOutputStream.Caches.subclassAudits);
        ObjectStreamClass.WeakClassKey weakClassKey = new ObjectStreamClass.WeakClassKey(class1, ObjectOutputStream.Caches.subclassAuditsQueue);
        Boolean value = (Boolean)ObjectOutputStream.Caches.subclassAudits.get(weakClassKey);
        if (value == null) {
            value = Boolean.valueOf(auditSubclass(class1));
            ObjectOutputStream.Caches.subclassAudits.putIfAbsent(weakClassKey, value);
        }
        if ((boolean)value) {
            return;
        }
        securityManager.checkPermission(ObjectOutputStream.SUBCLASS_IMPLEMENTATION_PERMISSION);
        return;
    }
    
    private static boolean auditSubclass(final Class<?> clazz) {
        return (boolean)AccessController.doPrivileged(new PrivilegedAction<Boolean>() {
            @Override
            public Boolean run() {
                Class<Object> clazz = clazz;
                while (clazz != ObjectOutputStream.class) {
                    try {
                        clazz.getDeclaredMethod("writeUnshared", new Class[] { Object.class });
                        return Boolean.FALSE;
                    }
                    catch (NoSuchMethodException ex) {
                        try {
                            clazz.getDeclaredMethod("putFields", (Class[])null);
                            return Boolean.FALSE;
                        }
                        catch (NoSuchMethodException ex2) {
                            clazz = clazz.getSuperclass();
                        }
                    }
                }
                return Boolean.TRUE;
            }
        });
    }
    
    private void clear() {
        this.subs.clear();
        this.handles.clear();
    }
    
    private void writeObject0(Object o, boolean b) throws IOException {
        int setBlockDataMode = this.bout.setBlockDataMode(false);
        ++this.depth;
        try {
            if ((o = this.subs.lookup(o)) == null) {
                this.writeNull();
                return;
            }
            int lookup;
            if (!b && (lookup = this.handles.lookup(o)) != -1) {
                this.writeHandle(lookup);
                return;
            }
            if ((o instanceof Class)) {
                this.writeClass((Class)o, b);
                return;
            }
            if ((o instanceof ObjectStreamClass)) {
                this.writeClassDesc((ObjectStreamClass)o, b);
                return;
            }
            Class<Object> clazz = o.getClass();
            ObjectStreamClass objectStreamClass;
            while (true) {
                objectStreamClass = ObjectStreamClass.lookup(clazz, true);
                Class<Object> class1;
                if (objectStreamClass.hasWriteReplaceMethod() && (o = objectStreamClass.invokeWriteReplace(o)) != null && (class1 = o.getClass()) != clazz) {
                    clazz = class1;
                }
                else {
                    break;
                }
            }
            if (this.enableReplace) {
                Object replaceObject = this.replaceObject(o);
                if (replaceObject != o && replaceObject != null) {
                    clazz = replaceObject.getClass();
                    objectStreamClass = ObjectStreamClass.lookup(clazz, true);
                }
                o = replaceObject;
            }
            if (o != o) {
                this.subs.assign(o, o);
                if (o == null) {
                    this.writeNull();
                    return;
                }
                int lookup2;
                if (!b && (lookup2 = this.handles.lookup(o)) != -1) {
                    this.writeHandle(lookup2);
                    return;
                }
                if ((o instanceof Class)) {
                    this.writeClass((Class)o, b);
                    return;
                }
                if ((o instanceof ObjectStreamClass)) {
                    this.writeClassDesc((ObjectStreamClass)o, b);
                    return;
                }
            }
            if ((o instanceof String)) {
                this.writeString((String)o, b);
            }
            else if (clazz.isArray()) {
                this.writeArray(o, objectStreamClass, b);
            }
            else if ((o instanceof Enum)) {
                this.writeEnum((Enum)o, objectStreamClass, b);
            }
            else if ((o instanceof Serializable)) {
                this.writeOrdinaryObject(o, objectStreamClass, b);
            }
            else {
                if (ObjectOutputStream.extendedDebugInfo) {
                    throw new NotSerializableException(clazz.getName() + "\n" + this.debugInfoStack.toString());
                }
                throw new NotSerializableException(clazz.getName());
            }
        }
        finally {
            --this.depth;
            this.bout.setBlockDataMode((boolean)setBlockDataMode);
        }
    }
    
    private void writeNull() throws IOException {
        this.bout.writeByte(112);
    }
    
    private void writeHandle(int n) throws IOException {
        this.bout.writeByte(113);
        this.bout.writeInt(8257536 + n);
    }
    
    private void writeClass(Class<?> clazz, boolean b) throws IOException {
        this.bout.writeByte(118);
        this.writeClassDesc(ObjectStreamClass.lookup(clazz, true), false);
        this.handles.assign(b ? null : clazz);
    }
    
    private void writeClassDesc(ObjectStreamClass objectStreamClass, boolean b) throws IOException {
        if (objectStreamClass == null) {
            this.writeNull();
        }
        else {
            int lookup;
            if (!b && (lookup = this.handles.lookup(objectStreamClass)) != -1) {
                this.writeHandle(lookup);
            }
            else if (objectStreamClass.isProxy()) {
                this.writeProxyDesc(objectStreamClass, b);
            }
            else {
                this.writeNonProxyDesc(objectStreamClass, b);
            }
        }
    }
    
    private boolean isCustomSubclass() {
        return this.getClass().getClassLoader() != ObjectOutputStream.class.getClassLoader();
    }
    
    private void writeProxyDesc(ObjectStreamClass objectStreamClass, boolean b) throws IOException {
        this.bout.writeByte(125);
        this.handles.assign(b ? null : objectStreamClass);
        Class<Object> forClass = objectStreamClass.forClass();
        Class[] interfaces = forClass.getInterfaces();
        this.bout.writeInt(interfaces.length);
        for (int i = 0; i < interfaces.length; ++i) {
            this.bout.writeUTF(interfaces[i].getName());
        }
        this.bout.setBlockDataMode(true);
        if (forClass != null && this.isCustomSubclass()) {
            ReflectUtil.checkPackageAccess(forClass);
        }
        this.annotateProxyClass(forClass);
        this.bout.setBlockDataMode(false);
        this.bout.writeByte(120);
        this.writeClassDesc(objectStreamClass.getSuperDesc(), false);
    }
    
    private void writeNonProxyDesc(ObjectStreamClass objectStreamClass, boolean b) throws IOException {
        this.bout.writeByte(114);
        this.handles.assign(b ? null : objectStreamClass);
        if (this.protocol == 1) {
            objectStreamClass.writeNonProxy(this);
        }
        else {
            this.writeClassDescriptor(objectStreamClass);
        }
        Class<Object> forClass = objectStreamClass.forClass();
        this.bout.setBlockDataMode(true);
        if (forClass != null && this.isCustomSubclass()) {
            ReflectUtil.checkPackageAccess(forClass);
        }
        this.annotateClass(forClass);
        this.bout.setBlockDataMode(false);
        this.bout.writeByte(120);
        this.writeClassDesc(objectStreamClass.getSuperDesc(), false);
    }
    
    private void writeString(String s, boolean b) throws IOException {
        this.handles.assign(b ? null : s);
        long utfLength = this.bout.getUTFLength(s);
        if (utfLength <= 65535L) {
            this.bout.writeByte(116);
            this.bout.writeUTF(s, utfLength);
        }
        else {
            this.bout.writeByte(124);
            this.bout.writeLongUTF(s, utfLength);
        }
    }
    
    private void writeArray(Object o, ObjectStreamClass objectStreamClass, boolean b) throws IOException {
        this.bout.writeByte(117);
        this.writeClassDesc(objectStreamClass, false);
        this.handles.assign(b ? null : o);
        Class<Object> componentType = objectStreamClass.forClass().getComponentType();
        if (componentType.isPrimitive()) {
            if (componentType == Integer.TYPE) {
                int[] array9 = (int[])(int[])o;
                this.bout.writeInt(array9.length);
                this.bout.writeInts(array9, 0, array9.length);
            }
            else if (componentType == Byte.TYPE) {
                byte[] array8 = (byte[])(byte[])o;
                this.bout.writeInt(array8.length);
                this.bout.write(array8, 0, array8.length, true);
            }
            else if (componentType == Long.TYPE) {
                long[] array7 = (long[])(long[])o;
                this.bout.writeInt(array7.length);
                this.bout.writeLongs(array7, 0, array7.length);
            }
            else if (componentType == Float.TYPE) {
                float[] array6 = (float[])(float[])o;
                this.bout.writeInt(array6.length);
                this.bout.writeFloats(array6, 0, array6.length);
            }
            else if (componentType == Double.TYPE) {
                double[] array5 = (double[])(double[])o;
                this.bout.writeInt(array5.length);
                this.bout.writeDoubles(array5, 0, array5.length);
            }
            else if (componentType == Short.TYPE) {
                short[] array4 = (short[])(short[])o;
                this.bout.writeInt(array4.length);
                this.bout.writeShorts(array4, 0, array4.length);
            }
            else if (componentType == Character.TYPE) {
                char[] array3 = (char[])(char[])o;
                this.bout.writeInt(array3.length);
                this.bout.writeChars(array3, 0, array3.length);
            }
            else if (componentType == Boolean.TYPE) {
                boolean[] array2 = (boolean[])(boolean[])o;
                this.bout.writeInt(array2.length);
                this.bout.writeBooleans(array2, 0, array2.length);
            }
            else {
                throw new InternalError();
            }
        }
        else {
            Object[] array = (Object[])(Object[])o;
            int length = array.length;
            this.bout.writeInt(length);
            if (ObjectOutputStream.extendedDebugInfo) {
                this.debugInfoStack.push("array (class \"" + o.getClass().getName() + "\", size: " + length + ")");
            }
            try {
                for (int i = 0; i < length; ++i) {
                    if (ObjectOutputStream.extendedDebugInfo) {
                        this.debugInfoStack.push("element of array (index: " + i + ")");
                    }
                    try {
                        this.writeObject0(array[i], false);
                    }
                    finally {}
                }
            }
            finally {
                if (ObjectOutputStream.extendedDebugInfo) {
                    this.debugInfoStack.pop();
                }
            }
        }
    }
    
    private void writeEnum(Enum<?> enum1, ObjectStreamClass objectStreamClass, boolean b) throws IOException {
        this.bout.writeByte(126);
        ObjectStreamClass superDesc = objectStreamClass.getSuperDesc();
        this.writeClassDesc((superDesc.forClass() == Enum.class) ? objectStreamClass : superDesc, false);
        this.handles.assign(b ? null : enum1);
        this.writeString(enum1.name(), false);
    }
    
    private void writeOrdinaryObject(Object o, ObjectStreamClass objectStreamClass, boolean b) throws IOException {
        if (ObjectOutputStream.extendedDebugInfo) {
            this.debugInfoStack.push(((this.depth == 1) ? "root " : "") + "object (class \"" + o.getClass().getName() + "\", " + o.toString() + ")");
        }
        try {
            objectStreamClass.checkSerialize();
            this.bout.writeByte(115);
            this.writeClassDesc(objectStreamClass, false);
            this.handles.assign(b ? null : o);
            if (objectStreamClass.isExternalizable() && !objectStreamClass.isProxy()) {
                this.writeExternalData((Externalizable)o);
            }
            else {
                this.writeSerialData(o, objectStreamClass);
            }
        }
        finally {
            if (ObjectOutputStream.extendedDebugInfo) {
                this.debugInfoStack.pop();
            }
        }
    }
    
    private void writeExternalData(Externalizable externalizable) throws IOException {
        ObjectOutputStream.PutFieldImpl curPut = this.curPut;
        this.curPut = null;
        if (ObjectOutputStream.extendedDebugInfo) {
            this.debugInfoStack.push("writeExternal data");
        }
        SerialCallbackContext curContext = this.curContext;
        try {
            this.curContext = null;
            if (this.protocol == 1) {
                externalizable.writeExternal(this);
            }
            else {
                this.bout.setBlockDataMode(true);
                externalizable.writeExternal(this);
                this.bout.setBlockDataMode(false);
                this.bout.writeByte(120);
            }
        }
        finally {
            this.curContext = curContext;
            if (ObjectOutputStream.extendedDebugInfo) {
                this.debugInfoStack.pop();
            }
        }
        this.curPut = curPut;
    }
    
    private void writeSerialData(Object o, ObjectStreamClass objectStreamClass) throws IOException {
        ObjectStreamClass.ClassDataSlot[] classDataLayout = objectStreamClass.getClassDataLayout();
        for (int i = 0; i < classDataLayout.length; ++i) {
            ObjectStreamClass desc = classDataLayout[i].desc;
            if (desc.hasWriteObjectMethod()) {
                ObjectOutputStream.PutFieldImpl curPut = this.curPut;
                this.curPut = null;
                SerialCallbackContext curContext = this.curContext;
                if (ObjectOutputStream.extendedDebugInfo) {
                    this.debugInfoStack.push("custom writeObject data (class \"" + desc.getName() + "\")");
                }
                try {
                    this.curContext = new SerialCallbackContext(o, desc);
                    this.bout.setBlockDataMode(true);
                    desc.invokeWriteObject(o, this);
                    this.bout.setBlockDataMode(false);
                    this.bout.writeByte(120);
                }
                finally {
                    this.curContext.setUsed();
                    this.curContext = curContext;
                    if (ObjectOutputStream.extendedDebugInfo) {
                        this.debugInfoStack.pop();
                    }
                }
                this.curPut = curPut;
            }
            else {
                this.defaultWriteFields(o, desc);
            }
        }
    }
    
    private void defaultWriteFields(Object o, ObjectStreamClass objectStreamClass) throws IOException {
        Class<Object> forClass = objectStreamClass.forClass();
        if (forClass != null && o != null && !forClass.isInstance(o)) {
            throw new ClassCastException();
        }
        objectStreamClass.checkDefaultSerialize();
        int primDataSize = objectStreamClass.getPrimDataSize();
        if (this.primVals == null || this.primVals.length < primDataSize) {
            this.primVals = new byte[primDataSize];
        }
        objectStreamClass.getPrimFieldValues(o, this.primVals);
        this.bout.write(this.primVals, 0, primDataSize, false);
        ObjectStreamField[] fields = objectStreamClass.getFields(false);
        Object[] array = new Object[objectStreamClass.getNumObjFields()];
        int n = fields.length - array.length;
        objectStreamClass.getObjFieldValues(o, array);
        for (int i = 0; i < array.length; ++i) {
            if (ObjectOutputStream.extendedDebugInfo) {
                this.debugInfoStack.push("field (class \"" + objectStreamClass.getName() + "\", name: \"" + fields[n + i].getName() + "\", type: \"" + fields[n + i].getType() + "\")");
            }
            try {
                this.writeObject0(array[i], fields[n + i].isUnshared());
            }
            finally {
                if (ObjectOutputStream.extendedDebugInfo) {
                    this.debugInfoStack.pop();
                }
            }
        }
        return;
    }
    
    private void writeFatalException(IOException ex) throws IOException {
        this.clear();
        int setBlockDataMode = this.bout.setBlockDataMode(false);
        try {
            this.bout.writeByte(123);
            this.writeObject0(ex, false);
            this.clear();
        }
        finally {
            this.bout.setBlockDataMode((boolean)setBlockDataMode);
        }
    }
    
    private static native void floatsToBytes(float[] p0, int p1, byte[] p2, int p3, int p4);
    
    private static native void doublesToBytes(double[] p0, int p1, byte[] p2, int p3, int p4);
    
    static void access$300(ObjectOutputStream objectOutputStream, Object o, boolean b) throws IOException {
        objectOutputStream.writeObject0(o, b);
    }
    
    static void access$400(float[] array, int n, byte[] array2, int n2, int n3) {
        floatsToBytes(array, n, array2, n2, n3);
    }
    
    static void access$500(double[] array, int n, byte[] array2, int n2, int n3) {
        doublesToBytes(array, n, array2, n2, n3);
    }
    
    static {
        extendedDebugInfo = (boolean)AccessController.doPrivileged(new GetBooleanAction("sun.io.serialization.extendedDebugInfo"));
    }
    
    private static class Caches
    {
        static final ConcurrentMap<ObjectStreamClass.WeakClassKey, Boolean> subclassAudits;
        static final ReferenceQueue<Class<?>> subclassAuditsQueue;
        
        static {
            subclassAudits = new ConcurrentHashMap();
            subclassAuditsQueue = new ReferenceQueue();
        }
    }
    
    public abstract static class PutField
    {
        public abstract void put(String p0, boolean p1);
        
        public abstract void put(String p0, byte p1);
        
        public abstract void put(String p0, char p1);
        
        public abstract void put(String p0, short p1);
        
        public abstract void put(String p0, int p1);
        
        public abstract void put(String p0, long p1);
        
        public abstract void put(String p0, float p1);
        
        public abstract void put(String p0, double p1);
        
        public abstract void put(String p0, Object p1);
        
        @Deprecated
        public abstract void write(ObjectOutput p0) throws IOException;
    }
    
    private class PutFieldImpl extends ObjectOutputStream.PutField
    {
        private final ObjectStreamClass desc;
        private final byte[] primVals;
        private final Object[] objVals;
        
        PutFieldImpl(ObjectStreamClass desc) {
            this.desc = desc;
            this.primVals = new byte[desc.getPrimDataSize()];
            this.objVals = new Object[desc.getNumObjFields()];
        }
        
        @Override
        public void put(String s, boolean b) {
            Bits.putBoolean(this.primVals, this.getFieldOffset(s, Boolean.TYPE), b);
        }
        
        @Override
        public void put(String s, byte b) {
            this.primVals[this.getFieldOffset(s, Byte.TYPE)] = b;
        }
        
        @Override
        public void put(String s, char c) {
            Bits.putChar(this.primVals, this.getFieldOffset(s, Character.TYPE), c);
        }
        
        @Override
        public void put(String s, short n) {
            Bits.putShort(this.primVals, this.getFieldOffset(s, Short.TYPE), n);
        }
        
        @Override
        public void put(String s, int n) {
            Bits.putInt(this.primVals, this.getFieldOffset(s, Integer.TYPE), n);
        }
        
        @Override
        public void put(String s, float n) {
            Bits.putFloat(this.primVals, this.getFieldOffset(s, Float.TYPE), n);
        }
        
        @Override
        public void put(String s, long n) {
            Bits.putLong(this.primVals, this.getFieldOffset(s, Long.TYPE), n);
        }
        
        @Override
        public void put(String s, double n) {
            Bits.putDouble(this.primVals, this.getFieldOffset(s, Double.TYPE), n);
        }
        
        @Override
        public void put(String s, Object o) {
            this.objVals[this.getFieldOffset(s, Object.class)] = o;
        }
        
        @Override
        public void write(ObjectOutput objectOutput) throws IOException {
            if (ObjectOutputStream.this != objectOutput) {
                throw new IllegalArgumentException("wrong stream");
            }
            objectOutput.write(this.primVals, 0, this.primVals.length);
            ObjectStreamField[] fields = this.desc.getFields(false);
            int n = fields.length - this.objVals.length;
            int i = 0;
            while (i < this.objVals.length) {
                if (fields[n + i].isUnshared()) {
                    throw new IOException("cannot write unshared object");
                }
                objectOutput.writeObject(this.objVals[i]);
                ++i;
            }
            return;
        }
        
        void writeFields() throws IOException {
            ObjectOutputStream.this.bout.write(this.primVals, 0, this.primVals.length, false);
            ObjectStreamField[] fields = this.desc.getFields(false);
            int n = fields.length - this.objVals.length;
            for (int i = 0; i < this.objVals.length; ++i) {
                if (ObjectOutputStream.extendedDebugInfo) {
                    ObjectOutputStream.this.debugInfoStack.push("field (class \"" + this.desc.getName() + "\", name: \"" + fields[n + i].getName() + "\", type: \"" + fields[n + i].getType() + "\")");
                }
                try {
                    ObjectOutputStream.access$300(ObjectOutputStream.this, this.objVals[i], fields[n + i].isUnshared());
                }
                finally {
                    if (ObjectOutputStream.extendedDebugInfo) {
                        ObjectOutputStream.this.debugInfoStack.pop();
                    }
                }
            }
        }
        
        private int getFieldOffset(String str, Class<?> obj) {
            ObjectStreamField field = this.desc.getField(str, obj);
            if (field == null) {
                throw new IllegalArgumentException("no such field " + str + " with type " + obj);
            }
            return field.getOffset();
        }
    }
    
    private static class BlockDataOutputStream extends OutputStream implements DataOutput
    {
        private static final int MAX_BLOCK_SIZE = 1024;
        private static final int MAX_HEADER_SIZE = 5;
        private static final int CHAR_BUF_SIZE = 256;
        private final byte[] buf;
        private final byte[] hbuf;
        private final char[] cbuf;
        private boolean blkmode;
        private int pos;
        private final OutputStream out;
        private final DataOutputStream dout;
        
        BlockDataOutputStream(OutputStream out) {
            this.buf = new byte[1024];
            this.hbuf = new byte[5];
            this.cbuf = new char[256];
            this.blkmode = false;
            this.pos = 0;
            this.out = out;
            this.dout = new DataOutputStream(this);
        }
        
        boolean setBlockDataMode(boolean blkmode) throws IOException {
            if (this.blkmode == blkmode) {
                return this.blkmode;
            }
            this.drain();
            this.blkmode = blkmode;
            return !this.blkmode;
        }
        
        boolean getBlockDataMode() {
            return this.blkmode;
        }
        
        @Override
        public void write(int n) throws IOException {
            if (this.pos >= 1024) {
                this.drain();
            }
            this.buf[this.pos++] = (byte)n;
        }
        
        @Override
        public void write(byte[] array) throws IOException {
            this.write(array, 0, array.length, false);
        }
        
        @Override
        public void write(byte[] array, int n, int n2) throws IOException {
            this.write(array, n, n2, false);
        }
        
        @Override
        public void flush() throws IOException {
            this.drain();
            this.out.flush();
        }
        
        @Override
        public void close() throws IOException {
            this.flush();
            this.out.close();
        }
        
        void write(byte[] array, int n, int i, boolean b) throws IOException {
            if (!b && !this.blkmode) {
                this.drain();
                this.out.write(array, n, i);
                return;
            }
            while (i > 0) {
                if (this.pos >= 1024) {
                    this.drain();
                }
                if (i >= 1024 && !b && this.pos == 0) {
                    this.writeBlockHeader(1024);
                    this.out.write(array, n, 1024);
                    n += 1024;
                    i -= 1024;
                }
                else {
                    int min = Math.min(i, 1024 - this.pos);
                    System.arraycopy(array, n, this.buf, this.pos, min);
                    this.pos += min;
                    n += min;
                    i -= min;
                }
            }
            return;
        }
        
        void drain() throws IOException {
            if (this.pos == 0) {
                return;
            }
            if (this.blkmode) {
                this.writeBlockHeader(this.pos);
            }
            this.out.write(this.buf, 0, this.pos);
            this.pos = 0;
            return;
        }
        
        private void writeBlockHeader(int n) throws IOException {
            if (n <= 255) {
                this.hbuf[0] = 119;
                this.hbuf[1] = (byte)n;
                this.out.write(this.hbuf, 0, 2);
            }
            else {
                this.hbuf[0] = 122;
                Bits.putInt(this.hbuf, 1, n);
                this.out.write(this.hbuf, 0, 5);
            }
        }
        
        @Override
        public void writeBoolean(boolean b) throws IOException {
            if (this.pos >= 1024) {
                this.drain();
            }
            Bits.putBoolean(this.buf, this.pos++, b);
        }
        
        @Override
        public void writeByte(int n) throws IOException {
            if (this.pos >= 1024) {
                this.drain();
            }
            this.buf[this.pos++] = (byte)n;
        }
        
        @Override
        public void writeChar(int n) throws IOException {
            if (this.pos + 2 <= 1024) {
                Bits.putChar(this.buf, this.pos, (char)n);
                this.pos += 2;
            }
            else {
                this.dout.writeChar(n);
            }
        }
        
        @Override
        public void writeShort(int n) throws IOException {
            if (this.pos + 2 <= 1024) {
                Bits.putShort(this.buf, this.pos, (short)n);
                this.pos += 2;
            }
            else {
                this.dout.writeShort(n);
            }
        }
        
        @Override
        public void writeInt(int n) throws IOException {
            if (this.pos + 4 <= 1024) {
                Bits.putInt(this.buf, this.pos, n);
                this.pos += 4;
            }
            else {
                this.dout.writeInt(n);
            }
        }
        
        @Override
        public void writeFloat(float n) throws IOException {
            if (this.pos + 4 <= 1024) {
                Bits.putFloat(this.buf, this.pos, n);
                this.pos += 4;
            }
            else {
                this.dout.writeFloat(n);
            }
        }
        
        @Override
        public void writeLong(long n) throws IOException {
            if (this.pos + 8 <= 1024) {
                Bits.putLong(this.buf, this.pos, n);
                this.pos += 8;
            }
            else {
                this.dout.writeLong(n);
            }
        }
        
        @Override
        public void writeDouble(double n) throws IOException {
            if (this.pos + 8 <= 1024) {
                Bits.putDouble(this.buf, this.pos, n);
                this.pos += 8;
            }
            else {
                this.dout.writeDouble(n);
            }
        }
        
        @Override
        public void writeBytes(String s) throws IOException {
            int length = s.length();
            int n = 0;
            int min = 0;
            int min2;
            for (int i = 0; i < length; i += min2) {
                if (n >= min) {
                    n = 0;
                    min = Math.min(length - i, 256);
                    s.getChars(i, i + min, this.cbuf, 0);
                }
                if (this.pos >= 1024) {
                    this.drain();
                }
                min2 = Math.min(min - n, 1024 - this.pos);
                while (this.pos < this.pos + min2) {
                    this.buf[this.pos++] = (byte)this.cbuf[n++];
                }
            }
        }
        
        @Override
        public void writeChars(String s) throws IOException {
            int min;
            for (int length = s.length(), i = 0; i < length; i += min) {
                min = Math.min(length - i, 256);
                s.getChars(i, i + min, this.cbuf, 0);
                this.writeChars(this.cbuf, 0, min);
            }
        }
        
        @Override
        public void writeUTF(String s) throws IOException {
            this.writeUTF(s, this.getUTFLength(s));
        }
        
        void writeBooleans(boolean[] array, int i, int n) throws IOException {
            int a = i + n;
            while (i < a) {
                if (this.pos >= 1024) {
                    this.drain();
                }
                while (i < Math.min(a, i + (1024 - this.pos))) {
                    Bits.putBoolean(this.buf, this.pos++, array[i++]);
                }
            }
        }
        
        void writeChars(char[] array, int i, int n) throws IOException {
            int n2 = 1022;
            int a = i + n;
            while (i < a) {
                if (this.pos <= n2) {
                    while (i < Math.min(a, i + (1024 - this.pos >> 1))) {
                        Bits.putChar(this.buf, this.pos, array[i++]);
                        this.pos += 2;
                    }
                }
                else {
                    this.dout.writeChar((int)array[i++]);
                }
            }
        }
        
        void writeShorts(short[] array, int i, int n) throws IOException {
            int n2 = 1022;
            int a = i + n;
            while (i < a) {
                if (this.pos <= n2) {
                    while (i < Math.min(a, i + (1024 - this.pos >> 1))) {
                        Bits.putShort(this.buf, this.pos, array[i++]);
                        this.pos += 2;
                    }
                }
                else {
                    this.dout.writeShort((int)array[i++]);
                }
            }
        }
        
        void writeInts(int[] array, int i, int n) throws IOException {
            int n2 = 1020;
            int a = i + n;
            while (i < a) {
                if (this.pos <= n2) {
                    while (i < Math.min(a, i + (1024 - this.pos >> 2))) {
                        Bits.putInt(this.buf, this.pos, array[i++]);
                        this.pos += 4;
                    }
                }
                else {
                    this.dout.writeInt(array[i++]);
                }
            }
        }
        
        void writeFloats(float[] array, int i, int n) throws IOException {
            int n2 = 1020;
            int n3 = i + n;
            while (i < n3) {
                if (this.pos <= n2) {
                    int min = Math.min(n3 - i, 1024 - this.pos >> 2);
                    ObjectOutputStream.access$400(array, i, this.buf, this.pos, min);
                    i += min;
                    this.pos += min << 2;
                }
                else {
                    this.dout.writeFloat(array[i++]);
                }
            }
        }
        
        void writeLongs(long[] array, int i, int n) throws IOException {
            int n2 = 1016;
            int a = i + n;
            while (i < a) {
                if (this.pos <= n2) {
                    while (i < Math.min(a, i + (1024 - this.pos >> 3))) {
                        Bits.putLong(this.buf, this.pos, array[i++]);
                        this.pos += 8;
                    }
                }
                else {
                    this.dout.writeLong(array[i++]);
                }
            }
        }
        
        void writeDoubles(double[] array, int i, int n) throws IOException {
            int n2 = 1016;
            int n3 = i + n;
            while (i < n3) {
                if (this.pos <= n2) {
                    int min = Math.min(n3 - i, 1024 - this.pos >> 3);
                    ObjectOutputStream.access$500(array, i, this.buf, this.pos, min);
                    i += min;
                    this.pos += min << 3;
                }
                else {
                    this.dout.writeDouble(array[i++]);
                }
            }
        }
        
        long getUTFLength(String s) {
            int length = s.length();
            long n = 0L;
            int min;
            for (int i = 0; i < length; i += min) {
                min = Math.min(length - i, 256);
                s.getChars(i, i + min, this.cbuf, 0);
                for (int j = 0; j < min; ++j) {
                    int n2 = this.cbuf[j];
                    if (n2 >= '\u0001' && n2 <= '\u007f') {
                        ++n;
                    }
                    else if (n2 > '߿') {
                        n += 3L;
                    }
                    else {
                        n += 2L;
                    }
                }
            }
            return n;
        }
        
        void writeUTF(String s, long n) throws IOException {
            if (n > 65535L) {
                throw new UTFDataFormatException();
            }
            this.writeShort((int)n);
            if (n == (long)s.length()) {
                this.writeBytes(s);
            }
            else {
                this.writeUTFBody(s);
            }
            return;
        }
        
        void writeLongUTF(String s) throws IOException {
            this.writeLongUTF(s, this.getUTFLength(s));
        }
        
        void writeLongUTF(String s, long n) throws IOException {
            this.writeLong(n);
            if (n == (long)s.length()) {
                this.writeBytes(s);
            }
            else {
                this.writeUTFBody(s);
            }
        }
        
        private void writeUTFBody(String s) throws IOException {
            int n = 1021;
            int min;
            for (int length = s.length(), i = 0; i < length; i += min) {
                min = Math.min(length - i, 256);
                s.getChars(i, i + min, this.cbuf, 0);
                for (int j = 0; j < min; ++j) {
                    int n2 = this.cbuf[j];
                    if (this.pos <= n) {
                        if (n2 <= '\u007f' && n2 != '\0') {
                            this.buf[this.pos++] = (byte)n2;
                        }
                        else if (n2 > '߿') {
                            this.buf[this.pos + 2] = (byte)(0x80 | (n2 >> 0 & 0x3F));
                            this.buf[this.pos + 1] = (byte)(0x80 | (n2 >> 6 & 0x3F));
                            this.buf[this.pos + 0] = (byte)(0xE0 | (n2 >> 12 & 0xF));
                            this.pos += 3;
                        }
                        else {
                            this.buf[this.pos + 1] = (byte)(0x80 | (n2 >> 0 & 0x3F));
                            this.buf[this.pos + 0] = (byte)(0xC0 | (n2 >> 6 & 0x1F));
                            this.pos += 2;
                        }
                    }
                    else if (n2 <= '\u007f' && n2 != '\0') {
                        this.write(n2);
                    }
                    else if (n2 > '߿') {
                        this.write(0xE0 | (n2 >> 12 & 0xF));
                        this.write(0x80 | (n2 >> 6 & 0x3F));
                        this.write(0x80 | (n2 >> 0 & 0x3F));
                    }
                    else {
                        this.write(0xC0 | (n2 >> 6 & 0x1F));
                        this.write(0x80 | (n2 >> 0 & 0x3F));
                    }
                }
            }
        }
    }
    
    private static class HandleTable
    {
        private int size;
        private int threshold;
        private final float loadFactor;
        private int[] spine;
        private int[] next;
        private Object[] objs;
        
        HandleTable(int n, float loadFactor) {
            this.loadFactor = loadFactor;
            this.spine = new int[n];
            this.next = new int[n];
            this.objs = new Object[n];
            this.threshold = (int)((float)n * loadFactor);
            this.clear();
        }
        
        int assign(Object o) {
            if (this.size >= this.next.length) {
                this.growEntries();
            }
            if (this.size >= this.threshold) {
                this.growSpine();
            }
            this.insert(o, this.size);
            return this.size++;
        }
        
        int lookup(Object o) {
            if (this.size == 0) {
                return -1;
            }
            int i = this.spine[this.hash(o) % this.spine.length];
            while (i >= 0) {
                if (this.objs[i] == o) {
                    return i;
                }
                i = this.next[i];
            }
            return -1;
        }
        
        void clear() {
            Arrays.fill(this.spine, -1);
            Arrays.fill(this.objs, 0, this.size, null);
            this.size = 0;
        }
        
        int size() {
            return this.size;
        }
        
        private void insert(Object o, int n) {
            int n2 = this.hash(o) % this.spine.length;
            this.objs[n] = o;
            this.next[n] = this.spine[n2];
            this.spine[n2] = n;
        }
        
        private void growSpine() {
            this.spine = new int[(this.spine.length << 1) + 1];
            this.threshold = (int)((float)this.spine.length * this.loadFactor);
            Arrays.fill(this.spine, -1);
            for (int i = 0; i < this.size; ++i) {
                this.insert(this.objs[i], i);
            }
        }
        
        private void growEntries() {
            int n = (this.next.length << 1) + 1;
            int[] next = new int[n];
            System.arraycopy(this.next, 0, next, 0, this.size);
            this.next = next;
            Object[] objs = new Object[n];
            System.arraycopy(this.objs, 0, objs, 0, this.size);
            this.objs = objs;
        }
        
        private int hash(Object o) {
            return System.identityHashCode(o) & Integer.MAX_VALUE;
        }
    }
    
    private static class ReplaceTable
    {
        private final ObjectOutputStream.HandleTable htab;
        private Object[] reps;
        
        ReplaceTable(int n, float n2) {
            this.htab = new ObjectOutputStream.HandleTable(n, n2);
            this.reps = new Object[n];
        }
        
        void assign(Object o, Object o2) {
            int i = this.htab.assign(o);
            while (i >= this.reps.length) {
                this.grow();
            }
            this.reps[i] = o2;
        }
        
        Object lookup(Object o) {
            int lookup = this.htab.lookup(o);
            return (lookup >= 0) ? this.reps[lookup] : o;
        }
        
        void clear() {
            Arrays.fill(this.reps, 0, this.htab.size(), null);
            this.htab.clear();
        }
        
        int size() {
            return this.htab.size();
        }
        
        private void grow() {
            Object[] reps = new Object[(this.reps.length << 1) + 1];
            System.arraycopy(this.reps, 0, reps, 0, this.reps.length);
            this.reps = reps;
        }
    }
    
    private static class DebugTraceInfoStack
    {
        private final List<String> stack;
        
        DebugTraceInfoStack() {
            this.stack = new ArrayList();
        }
        
        void clear() {
            this.stack.clear();
        }
        
        void pop() {
            this.stack.remove(this.stack.size() - 1);
        }
        
        void push(String str) {
            this.stack.add("\t- " + str);
        }
        
        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            if (!this.stack.isEmpty()) {
                for (int i = this.stack.size(); i > 0; --i) {
                    sb.append((String)this.stack.get(i - 1) + ((i != 1) ? "\n" : ""));
                }
            }
            return sb.toString();
        }
    }
}
